// engine.cc: robot engine implementation.
// 
// Copyright (C) 2004 Frank Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
// 
// $Id: main.cc,v 2.7 2003/09/01 12:30:26 scmfcl Exp $

#include "config.h"

RCSID (engine, "$Id$");

#include "version.hxx"
#include "list.hxx"
#include "xstring.hxx"

#include "agent.hxx"
#include "world.hxx"
#include "engine.hxx"

Engine *Engine::engine = NULL;

GLfloat Engine::light_diffuse[]   = {.6, .6, .6, 1.0};
GLfloat Engine::light_ambient[]   = {0.0, 0.0, 0.0, 1.0};
GLfloat Engine::light_specular[]  = {.6, .6, .6, 1.0};
GLfloat Engine::light_attenuation = .6;
GLfloat Engine::light_direction[] = {0.0, 0.0, -1.0};
GLfloat Engine::light_pos[] = {0.0, 0.0, 0.0, 1.0};

void display () {
  assert (Engine::engine);
  Engine::engine->edisplay ();
  return;
}

void reshape (int w, int h) {
  assert (Engine::engine);
  Engine::engine->ereshape (w, h);
  return;
}

void idle () {
  assert (Engine::engine);
  Engine::engine->eidle ();
  return;
}

void mouse (int btn, int state, int x, int y) {
  assert (Engine::engine);
  Engine::engine->emouse (btn, state, x, y);
  return;
}

void motion (int x, int y) {
  assert (Engine::engine);
  Engine::engine->emotion (x, y);
  return;
}

void keyboard (unsigned char key, int x, int y) {
  assert (Engine::engine);
  Engine::engine->ekeyboard (key, x, y);
  return;
}

Engine::Engine (World *w, SimpleList<string> *an,
		double spos, double sangle) {
  assert (Engine::engine == NULL);
  Engine::engine = this;
  wrld = w;
  agents = new SimpleList<Agent>;
  unsigned short l = 0;
  string *str = an->first ();
  while (str) {
    agents->add_last (new Agent (l++, *str, wrld, agents));
    str = an->next ();
  }
  if (wrld->is_game ()) {
    assert (agents->length () == 2);
  }
  sens_pos = spos;
  sens_angle = sangle;
  eyex = eyey = -6.0;
  eyez = 10.0;
  phi = 45.0 * M_PI / 180.00;
  theta = -25.0 * M_PI / 180.00;
  run_cont = false;
  xpos = -1;
  ypos = -1;
  pointerId = 0;
  light_global_ambient[0] = light_global_ambient[1] = 
    light_global_ambient[2] = 0.1;
  light_global_ambient[3] = 1.0;
  steps = 0UL;
}

Engine::~Engine () {
  delete wrld;
  delete agents;
}

void Engine::run () {

  glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);

  glutInitWindowSize (500, 500);
  glutCreateWindow ("CyBot World");
  
  glutDisplayFunc (display);
  glutReshapeFunc (reshape);
  glutIdleFunc (idle);
  glutMouseFunc (mouse);
  glutMotionFunc (motion);
  glutKeyboardFunc (keyboard);

  glClearColor (0.4, 0.4, 0.4, 1.0);

  glShadeModel (GL_SMOOTH);
  glPolygonMode (GL_FRONT, GL_FILL);

  glEnable (GL_DEPTH_TEST);
  glDepthFunc (GL_LEQUAL);
  glClearDepth (1.0);

  glEnable (GL_NORMALIZE);

  glLightModeli (GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
  glLightModeli (GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);

  glLightModelfv (GL_LIGHT_MODEL_AMBIENT, light_global_ambient);

  glEnable (GL_LIGHTING);
  glEnable (GL_LIGHT0);
  glLightfv (GL_LIGHT0, GL_AMBIENT, light_ambient);
  glLightfv (GL_LIGHT0, GL_DIFFUSE, light_diffuse);
  glLightfv (GL_LIGHT0, GL_SPECULAR, light_specular);
  glLightfv (GL_LIGHT0, GL_CONSTANT_ATTENUATION, &light_attenuation);
  glLightfv (GL_LIGHT0, GL_SPOT_DIRECTION, light_direction);
  glLightf (GL_LIGHT0, GL_SPOT_CUTOFF, 45.0);
  glLightf (GL_LIGHT0, GL_SPOT_EXPONENT, 3.0);

  Object::setup ();
  wrld->init ();
  if (!wrld->is_game ()) {
    Agent *a = agents->first ();
    while (a) {
      a->init ();
      (*wrld) (a->positionX (), a->positionY ()) -> add_last (a);
    }
    a = agents->next ();
  }

  display ();

  glutMainLoop ();
}

void Engine::edisplay ()
{
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();

  glLightfv (GL_LIGHT0, GL_POSITION, light_pos);

  gluLookAt (eyex, eyey, eyez,
	     eyex + sin (phi) * cos (theta),
	     eyey + cos (phi) * cos (theta),
	     eyez + sin (theta),
	     0.0, 0.0, 1.0);

  wrld->render ();

  glFlush ();
  glutSwapBuffers ();
}

void Engine::ereshape (int w, int h)
{
  glViewport (0, 0, w, h);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective (70.0, h/w, 0.5, 500.0);
  return;
}

void Engine::eidle ()
{
  if (run_cont)
    step ();
  return;
}

void Engine::emouse (int btn, int state, int x, int y)
{
  if (btn == GLUT_LEFT_BUTTON) {
    if (state == GLUT_DOWN) {
      pointerId = glutGet (GLUT_WINDOW_CURSOR);
      xpos = x;
      ypos = y;
      glutSetCursor (GLUT_CURSOR_NONE);
      glutWarpPointer (xpos, ypos);
    } else {
      xpos = -1;
      glutSetCursor (pointerId);
    }
  }
  return;
}

void Engine::emotion (int x, int y)
{
  if (xpos < 0)
    return;

  double dx = x - xpos;
  double dy = y - ypos;

  if (dx == 0 && dy == 0)
    return;

  glutWarpPointer (xpos, ypos);

  phi += dx * sens_angle;
  if (phi > 2.0 * M_PI)
    phi -= 2.0 * M_PI;
  else if (phi < 0)
    phi += 2.0 * M_PI;
  theta += dy * sens_angle;
  if (theta < -M_PI_2 + M_PI / 180.0)
    theta = -M_PI_2+ M_PI / 180.0;
  else if (theta > M_PI_2 - M_PI / 180.0)
    theta = M_PI_2+ M_PI / 180.0;

  glutPostRedisplay ();

  return;
}

void Engine::ekeyboard (unsigned char key, int x, int y)
{
  GLfloat dx = sin (phi) * sens_pos;
  GLfloat dy = cos (phi) * sens_pos;
  switch (key) {
  case 's':
    eyex -= dx;
    eyey -= dy;
    break;
  case 'w':
    eyex += dx;
    eyey += dy;
    break;
  case 'a':
    eyex += -dy;
    eyey += dx;
    break;
  case 'd':
    eyex -= -dy;
    eyey -= dx;
    break;
  case 'q':
    eyez += sens_pos;
    break;
  case 'e':
    eyez -= sens_pos;
    break;
  case 'r':
    run_cont = !run_cont;
    break;
  case 'x':
    if (!run_cont)
      step ();
    break;
  case 27:
    agents->purge ();
    delete agents;
    delete wrld;
    exit (0);
  }
  glutPostRedisplay ();
  return;
}

void Engine::step ()
{
  if (agents->length () == 0) {
    run_cont = false;
    return;
  }

  if (wrld->is_game ()) {
    Agent *a = agents->first ();
    Agent *ra;
    if ((steps%2) == 0) {
      ra = a->turn (wrld);
    } else {
      a = agents->next ();
      ra = a->turn (wrld);
    }
    if (ra != a) {
      agents->purge ();
    }
    steps++;
  } else {
    Agent *a = agents->first ();
    while (a) {
      Agent *zap = a->turn (wrld);
      if (zap) {
	if (zap == a) {
	  Agent *aa = agents->next ();
	  agents->del (a);
	  a = aa;
	} else {
	  agents->del (zap);
	  a = agents->next ();
	}
      } else {
	a = agents->next ();
      }
    }
  }
  glutPostRedisplay ();
  return;
}

/// Local Variables:        ***
/// compile-command: "make" ***
/// End:                    ***
